from xlsxwriter.workbook import Workbook
from xlsxwriter.exceptions import FileCreateError
from collections import defaultdict
import sys
import xlrd
import re
import os 
import time 
import math
from .os_functions import choose_file,choose_sheet_column,get_require_files

import warnings 
warnings.filterwarnings('ignore')

def replace_re_special(word):
	#注意\要写在前面，因为后面循环替换了\\进去
	for special_symbol in r'\-+()[]{}.*^$~|?':
		new_special_symbol = '\\' + special_symbol
		word = word.replace(special_symbol, new_special_symbol)
	return word

def get_keyword_dict(path):
	found_tag = 0 
	while found_tag <= 0:
		try:
			wb = xlrd.open_workbook(path)
			found_tag += 1 
		except FileNotFoundError:
			input('\n "Keywords.xlsx" File Not Found in Current Folder!\n \
Please put the Keywords.xlsx to current folder and Press Enter to continue')
			continue 

	ws = wb.sheet_by_index(0)

	#保存每个关键词列的文字
	keyword_dict = defaultdict(set)

	#表头,根据表头获取应该写入红色还是蓝色，还是粗体
	header_list = []
	for x in ws.row(0):
		if type(x.value) == str and x.value.strip() != '':
			header = x.value.strip()
		elif (type(x.value) == float or type(x.value) == int) :
			header = str(x.value).rstrip('0').rstrip('.').strip()
		else:
			#为了防止两列中间隔一个空的表头单元格
			header = None
		header_list.append(header)

	if not header_list:
		print('Keywords.xlsx should have table header/headers such as red,blue,bold etc.')
		input('\nPress Enter to exit')
		sys.exit()

	seen_keywords = set()
	for row in list(ws.get_rows())[1:]:
		for i,format_word in enumerate(header_list):
			if format_word != None:
				keyword_value = row[i].value 
				if type(keyword_value) == float and math.ceil(keyword_value) == keyword_value:
					keyword = str(keyword_value).rstrip('0').rstrip('.').strip()
				else:  #必须去掉容易导致歧义的特殊符号
					keyword = replace_re_special(str(keyword_value).strip().lower())

				if keyword not in seen_keywords and keyword != "" :
					keyword_dict[format_word].add(keyword)
					seen_keywords.add(keyword)

	wb.release_resources()

	return keyword_dict

def get_rich_param_list(text,keyword_dict,insert_dict):
	"""
	通过文本和关键词获取需要填入xlsxwriter write_rich_string的第三个内容/param的格式,忽略大小写,re.I
	:param text:
	:param keyword_list 
	"""
	#关键词列表为空返回原文
	keyword_list = []
	for x in keyword_dict.values():
		keyword_list += x

	if not keyword_list:
		return {'record_list':text}

	#3.5 导致无法正确定位数字和文字
	keyword_list = [str(x).lower().strip() for x in  keyword_list ]
	keyword_list = sorted(keyword_list,key=lambda x:len(x),reverse=True)

	keyword_pat = u'('+ '|'.join(keyword_list) + ')'
	#通过关键词拆分的句子
	try:
		split_list = re.split(keyword_pat,text,flags=re.I)
		split_list =  [str(x) for x in split_list if x != '' ]	
		#找到的关键词目标
		findall_list = re.findall(keyword_pat,text,flags=re.I)
		findall_list = [str(x) for x in findall_list if x != '']
		#不符合re的匹配格式,直接返回原文本
	except TypeError:
		return {'record_list':text}
	#找不到目标也返回原来的文本
	if findall_list == []:
		return {'record_list':text}

	#需要保证所有都是字符串格式
	record_list = []

	#记录不同颜色的关键词列表
	keyword_record_dict = defaultdict(list)
	for k in keyword_dict:
		keyword_record_dict[k] = []

	while findall_list:
		keyword = findall_list.pop(0)
		target_index = split_list.index(keyword)

		#配合keyword_dict的小写
		#keyword = keyword.lower()
		for k in keyword_dict.keys():
			if replace_re_special(keyword.lower()) in keyword_dict[k]:
				keyword_record_dict[k].append(keyword)
				#没有出现过的关键字才能继续写入，否则不做写入
				split_list.insert(target_index,insert_dict[k.lower()])

		#按顺序记录
		record_list += split_list[0:target_index+2]
		split_list = split_list[target_index+2:]
	#记录最后一段,如果split_list 为空
	record_list += split_list
	#返回一个需要插入文字的record_list和一个命中关键词的keyword_record_dict {'red':['blueteeth','face']}

	return {'record_list':record_list,'keyword_record_dict':keyword_record_dict}


def write_header(xlsxwriter_ws,header_column):
	for column_index,c in enumerate(header_column):
		xlsxwriter_ws.write(0,column_index,c)
	return xlsxwriter_ws


def get_insert_dict(xlsxwriter_wb,keyword_dict):
	insert_dict = defaultdict()
	#高亮关键字格式  红色
	for format_word, k_list in keyword_dict.items():
		#区分颜色和字体粗体或斜体
		format_word = format_word.lower()
		if format_word == 'bold' :
			custom_format = xlsxwriter_wb.add_format({'bold':True})
		else:
			custom_format = xlsxwriter_wb.add_format({'font_color':format_word})
		insert_dict[format_word] = custom_format

	return insert_dict

def write2wb(xlsxwriter_wb,highlight_column_list,keyword_dict,highlight_column='Text'):

	#获取insert_dict 
	insert_dict = get_insert_dict(xlsxwriter_wb,keyword_dict)

	new_highlight_name ='{0} With Macro'.format(highlight_column)

	bold_format = xlsxwriter_wb.add_format({'bold':True})
	#第一个sheet记录高亮关键词结果
	xlsxwriter_ws = xlsxwriter_wb.add_worksheet('highlight')
	xlsxwriter_ws_target = xlsxwriter_wb.add_worksheet('keyword_matched')
	#第一个sheet表头，只写一列即可,表头标黄色
	xlsxwriter_ws.write(0,0,new_highlight_name,bold_format)

	#第二个sheet表头
	for i,k in enumerate(keyword_dict.keys()):
		xlsxwriter_ws_target.write(0,i,k,bold_format)

	total_len = len(highlight_column_list)
	
	print('')
	for row_index,value in enumerate(highlight_column_list):  #[7743:7745]
		if  (row_index + 1) % 2000 == 0 or (row_index + 1== total_len)  :
			print(' Processing...',row_index+1,'/',total_len)

		rich_param_dict = get_rich_param_list(value,keyword_dict,insert_dict)

		params = rich_param_dict['record_list']
		if type(params) == list:
			params = [ x for x in params if x != ''] 

		keyword_record_dict = rich_param_dict.get('keyword_record_dict',None)

		#1.写入第一个文档的部分
		column_index = 0
		if type(params) == str:
			xlsxwriter_ws.write(row_index+1,column_index,params)
		elif type(params) == list and params:  #是列表格式并且不是空列表 
			success = xlsxwriter_ws.write_rich_string(row_index+1,column_index,*params)	
			#如果返回是-5代表列表只有两个元素，不能写入,需要把文本取出来重新写
			if success == -5:
				one_word_text = params[-1]
				success = xlsxwriter_ws.write(row_index+1,column_index,one_word_text,params[0])
		else:
			xlsxwriter_ws.write(row_index+1,column_index,params)

		#写入第二个sheet的部分，通过关键词字典提取目标关键词
		if keyword_record_dict:
			counter = 0 
			for k,v in keyword_record_dict.items():
				if k == 'bold':
					cell_format = xlsxwriter_wb.add_format({'bold':True})
				else:
					cell_format = xlsxwriter_wb.add_format({'font_color':k})
				column_index = list(keyword_record_dict.keys()).index(k)

				v_text = ','.join(v)
				xlsxwriter_ws_target.write(row_index+1,column_index,v_text,cell_format)

	#调整每个列的宽度
	xlsxwriter_ws.set_column(0,0,width = 50)
	xlsxwriter_ws_target.set_column(0,len(keyword_dict),width=20)

	return xlsxwriter_wb

def highlight(xlsxwriter_wb,xlrd_ws,highlight_column,keyword_dict):
	"""
	可以自己选处理哪个内容
	"""
	#获取columns
	columns = [x.value for x in xlrd_ws.row(0)]
	#定位第几列需要高亮
	highlight_column_index = columns.index(highlight_column)
	#不取第一行
	highlight_column_series = [x.value for x in xlrd_ws.col(highlight_column_index)][1:]
	xlsxwriter_wb = write2wb(xlsxwriter_wb,highlight_column_series,keyword_dict,highlight_column)

	return xlsxwriter_wb

def save_xlsxwriter_wb(xlsxwriter_wb,path):
	close_tag = 0
	while close_tag <= 0:
		try:
			xlsxwriter_wb.close()
			close_tag += 1 
		except FileCreateError:
			input('\nFailed to write file!\n  Please Close "{}" Then Press Enter to Continue'.format(path))
	print('\n{0} Saved'.format(path))


def highlight_kw(path,keyword_dict):

	content_path = choose_file(path)
	xlrd_wb,xlrd_ws,sheet_name,highlight_column = choose_sheet_column(content_path)

	new_path = re.match('(.*).(xlsx|xls|xlsm)$',content_path).group(1) + '_Macro.xlsx'

	xlsxwriter_wb = Workbook(new_path)
	xlsxwriter_wb = highlight(xlsxwriter_wb,xlrd_ws,highlight_column,keyword_dict)

	#传过来只是为了关闭
	xlrd_wb.release_resources()

	save_xlsxwriter_wb(xlsxwriter_wb,new_path)

if __name__ == '__main__' :
	keyword_path = get_require_files('.\\',["keyword"],if_walk_path=False)['keyword']

	keyword_dict = get_keyword_dict(keyword_path)

	highlight_kw('.\\',keyword_dict)

	input("\nPress Enter to Exit")



from .os_functions import enter_exit 
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool
from sqlalchemy.sql import text
from pathlib import Path
from configparser import ConfigParser
from shutil import copyfile 
import pandas as pd 
import traceback
import logging
import os 
import cx_Oracle 

# 批量扫描需要增加的表是否都有数据且数据正常更新

def get_connection_by_config(database_type='mysql',config_path= 'connection_config.ini'):
    cfparser = ConfigParser()
    cfparser.read('connection_config.ini',encoding="utf-8")
    config_sections = cfparser.sections()
    print('读取到的Sections:',config_sections)

    host = cfparser.get(database_type,'host')
    port = cfparser.get(database_type,'port')
    database = cfparser.get(database_type,'database')
    charset = cfparser.get(database_type,'charset')
    username = cfparser.get(database_type,'username')
    password = cfparser.get(database_type,'password')

    if database_type == 'mssql':
        database_type += '+pymssql'

    engine_text = f'{database_type}://{username}:{password}@{host}:{port}/{database}?charset={charset}'

    conn, db = get_sql_connection(engine_text)

    return conn, db

def gen_engine_text(config_path, database_type):
    #读取ini配置
    cfparser = ConfigParser()
    cfparser.read(config_path,encoding="utf-8")
    config_sections = cfparser.sections()

    host = cfparser.get(database_type,'host')
    port = cfparser.get(database_type,'port')
    database = cfparser.get(database_type,'database')
    charset = cfparser.get(database_type,'charset')
    username = cfparser.get(database_type,'username')
    password = cfparser.get(database_type,'password')

    engine_text = f'{database_type}://{username}:{password}@{host}:{port}/{database}?charset={charset}'

    if 'oracle' in database_type.lower():
        #Oracle需要指定client位置
        lib_dir = cfparser.get(database_type,'oci_folder_path')
        cx_Oracle.init_oracle_client(lib_dir=lib_dir)

        database = cfparser.get(database_type,'database')
        try:
            service_name = cfparser.get(database_type,'service_name')
        except:
            service_name = ''

        try:
            sid = cfparser.get(database_type,'sid')
        except:
            sid = ''

    # engine = create_engine("oracle+cx_oracle://<username>:<password>@(DESCRIPTION = (ADDRESS = (PROTOCOL = TCP)(HOST = <host>)(PORT = 1521)) (CONNECT_DATA = (SERVER = DEDICATED) (SERVICE_NAME = devdb)))") 

    if 'oracle' in database_type.lower() and service_name != '':
        engine_text = f'{database_type}://{username}:{password}@(DESCRIPTION = (ADDRESS = (PROTOCOL = TCP)(HOST = {host})(PORT = 1521)) (CONNECT_DATA = (SERVER = DEDICATED) (schema={database}) (SERVICE_NAME = {service_name}) (CHARSET=utf8)))'

    if 'oracle' in database_type.lower() and sid != '':
        engine_text = f'{database_type}://{username}:{password}@(DESCRIPTION = (ADDRESS = (PROTOCOL = TCP)(HOST = {host})(PORT = 1521)) (CONNECT_DATA = (SERVER = DEDICATED) (schema={database})(SID = {sid}) (CHARSET=utf8)))'

    return engine_text


def get_sql_connection(engine_text):
    db = create_engine(engine_text,poolclass= NullPool)
    conn = db.connect()
    return conn, db

def close_sql_connection(conn,db):
    conn.close()
    db.dispose()
    
def execute_fetchall(conn,sql):
    try:
        sql =  text(sql)
        result = conn.execute(sql).fetchall()
        return result
    except :
        print('没有找到任何对应数据或者未知异常\n SQL:')
        print(sql)
        return None 
        
def convert_fetchall2df(fetchall_result):
    if fetchall_result :
        return pd.DataFrame(fetchall_result,columns=fetchall_result[0].keys())
    else:
        print('没有转换成功DF数据或未知异常')
        return pd.DataFrame([]) 


def execute_fetchall_engine(engine_text,sql):
    # print(engine_text)
    db = create_engine(engine_text,poolclass= NullPool)
    conn = db.connect()
    result = None 
    sql =  text(sql)
    try:
        fetchall_result = conn.execute(sql).fetchall()
        if fetchall_result :
            print('Results get:{} rows'.format(len(fetchall_result)))
            result = pd.DataFrame(fetchall_result,columns=fetchall_result[0].keys())
        else:
            print('Results set is empty')
            return pd.DataFrame([])
    except Exception as e :
        print('No data was found or unknow error happened\n:')
        logging.error(traceback.format_exc(e))
        return pd.DataFrame([])
    finally:
        conn.close()
        db.dispose()

    return result 

def execute_updates_engine(engine_text,sql):
    #通过engine_text直接执行语句
    conn, db = get_sql_connection(engine_text)
    conn.execute(" set sql_safe_updates = 0; ")
    try:
        conn.execute(sql)
    except Exception as e:
        print(sql,'\n执行失败')
        logging.error(traceback.format_exc(e))
    finally:
        close_sql_connection(conn,db)

def get_sql_result(conn,sql):
    result = execute_fetchall(conn,sql)
    result_df = convert_fetchall2df(result)

    return result_df 

def get_alter_index_text(conn,table_name):
    #获取到增加或删除表索引的可执行TEXT文本, 废弃的function,mysql load方法不用去掉索引写入也非常快
    show_index_sql = text("""
                      SELECT table_name AS `Table`,
                           index_name AS `Index`,
                           GROUP_CONCAT(column_name ORDER BY index_name,seq_in_index  ) AS `Columns`
                        FROM information_schema.statistics
                        WHERE table_schema = database()
                        and table_name  = '{0}'
                        GROUP BY table_name, index_name
                        order by max(seq_in_index), length(GROUP_CONCAT(column_name ORDER BY index_name,seq_in_index )) 
                    """.format(table_name)) 
    #将获取的索引结果组合，清除掉索引 
    result = conn.execute(show_index_sql)
    table_indexes = result.fetchall()

    add_index_list = [ ]
    del_index_list = [ ]

    for idx in table_indexes:
        index_name = idx[1]
        column_name = idx[2]

        if index_name.upper() != 'PRIMARY':
            add_index_text = 'add index ' + index_name + '(' + column_name + ')' + '\n'
            del_index_text = 'drop index ' + index_name  + '\n'

            add_index_list.append(add_index_text)
            del_index_list.append(del_index_text)

    if add_index_list:
        add_index_text = text('alter table {} '.format(table_name) + ','.join(add_index_list) + ';').execution_options(autocommit=True)
        del_index_text = text('alter table {} '.format(table_name) + ','.join(del_index_list) + ';').execution_options(autocommit=True)

    return add_index_text,del_index_text 

def write2table(engine_text,df,table_name,how='normal'):
    # engine_text format : "mysql://root:00000000@localhost:3306/web_data?charset=utf8"
    """
    4种方式写入: 1.normal: 直接写入，可选参数是否清空原有表；
                2.complete_rewrite: 删除原有的所有数据，并整个写入新数据
                3.mysql_load：直接通过mysql_load 方式写入 (这个方式经测试已经非常快，不用再重建索引也很快)
               （适用于特别大的数据集, 都需要确保MYSQL已经有完整的表结构）
    """
    db = create_engine(engine_text,poolclass=NullPool)
    conn = db.connect()

    #schema是发现oracle无法写入数据加上的
    if how == 'normal' :
        df.to_sql(table_name,con=conn,if_exists='append',index=False,chunksize=100000)

    elif how == 'complete_rewrite':
        try:
            truncate_statement = "truncate {};".format(table_name)
            conn.execute(truncate_statement)
        except :
            print('truncate table"{}" failed!'.format(table_name))

        df.to_sql(table_name,con=conn,if_exists='append',index=False,chunksize=100000)

    #如果采用第二种方式写入
    elif how == 'mysql_load' : 
        #如果存在自增主键,去掉表头的自增主键
        auto_increment_key = ''

        header_column_sql = text(" describe {} ;".format(table_name))
        result = conn.execute(header_column_sql).fetchall()
        if result:
            header_columns = [ x[0] for x in list(result) ]
        else:
            enter_exit('找不到表格:',table_name)
        #根据MYSQL表头结构构建一个能LOAD的CSV文档,如果表格存在自增主键，
        auto_increment_key_sql = """ SELECT COLUMN_NAME
                                     FROM INFORMATION_SCHEMA.COLUMNS  
                                     WHERE TABLE_SCHEMA = DATABASE()  AND TABLE_NAME = '{}'  AND DATA_TYPE = 'int'
                                     AND COLUMN_DEFAULT IS NULL AND IS_NULLABLE = 'NO' AND EXTRA like '%auto_increment%';
                                 """.format(table_name)

        auto_increment_key_sql = text(auto_increment_key_sql)
        result_auto = conn.execute(auto_increment_key_sql).fetchone()

        if result_auto != None:
            auto_increment_key = list(result_auto)[0]

        header_columns = [x for x in header_columns if x != auto_increment_key]

        for h in header_columns:
            if h not in df.columns:
                df[h] = None

        df = df.loc[:,header_columns]
        #保存数据到临时目的地
        temp_path = get_most_upper_level_path('df_temp_file.csv')

        df.to_csv(temp_path,encoding='utf8',sep=',',quotechar='"',escapechar='\\',index=False,header=None)

        if auto_increment_key != '':
            load_infile_sql = r"""  LOAD DATA INFILE '{0}'
                                    INTO TABLE {1} 
                                    CHARACTER SET 'utf8mb4'
                                    FIELDS TERMINATED BY ',' ENCLOSED BY '"'
                                    ESCAPED BY '\\'
                                    LINES TERMINATED BY '\r\n'
                                    ({2})
                                    SET {3} = NULL; -- 将默认需要自增的键设置为NULL即可正常写入
                                    """.format(temp_path,table_name,','.join(header_columns),auto_increment_key)
        else:
            load_infile_sql = r"""  LOAD DATA INFILE '{0}'
                        INTO TABLE {1} 
                        CHARACTER SET 'utf8mb4'
                        FIELDS TERMINATED BY ',' ENCLOSED BY '"'
                        ESCAPED BY '\\'
                        LINES TERMINATED BY '\r\n';
                        """.format(temp_path,table_name)

        load_infile_sql = text(load_infile_sql).execution_options(autocommit=True)
        try:
            conn.execute(load_infile_sql)
        except Exception as e:
            logging.error(traceback.format_exc())
        finally:
            conn.close()
            db.dispose()
        #删除temp CSV文档
        os.remove(temp_path)
    else:
        print(how,'写入方法未知,数据未写入')

    conn.close()
    db.dispose()

def load_sql_data(engine_text,table_name,load_file_path,if_truncate=False):
    """write csv file into table"""
    db = create_engine(engine_text,poolclass=NullPool)
    conn = db.connect()

    # temp_path = get_most_upper_level_path('df_temp_file.csv')


    #复制一份数据到英文路径
    # copyfile(load_file_path,temp_path)

    load_infile_sql = r"""  LOAD DATA INFILE '{0}'
            INTO TABLE {1} 
            CHARACTER SET 'utf8mb4'
            FIELDS TERMINATED BY ',' ENCLOSED BY '"'
            ESCAPED BY '\\'
            LINES TERMINATED BY '\r\n';
            """.format(load_file_path,table_name)
    
    load_infile_sql = text(load_infile_sql)
    conn.execute(load_infile_sql)

    #删除临时文件
    # os.remove(temp_path)

    print(load_file_path,'dump complete.')

    conn.close()
    db.dispose()


def get_sql_data(engine_text,table_name,sql,save_path,how='normal'):
    """通过SQL获取到目标数据并保存到文档"""
    db = create_engine(engine_text,poolclass=NullPool)
    conn = db.connect()

    if how == 'normal':
        #execute 返回ResultProxy,fetchall如果为空返回空列表
        fetchall_result = execute_fetchall(conn,sql)
        result_df = convert_fetchall2df(fetchall_result)
        result_df.to_excel(save_path,index=False)
    elif how == 'mysql_dump':
         #保存数据到临时目的地,mysqldump不能识别中文的路径，所以必须获取到最高一级的类似C盘C://的路径才能正常写入
        temp_path = get_most_upper_level_path('df_temp_file.csv')

        #先获取一次表头,sql输入结尾不能填分号
        if ';' ==  sql.strip()[-1]:
            sql = sql.strip()[:-1]

        try:
            first_row = conn.execute(sql)
            header = list(first_row.keys())
        except Exception as e:
            logging.error(traceback.format_exc())
            print(sql,'数据提取失败')
            conn.close()
            db.dispose()
            return None

        dump_sql = r"""  {}
                        INTO OUTFILE '{}' 
                        CHARACTER SET 'utf8mb4'
                        FIELDS TERMINATED BY ',' ENCLOSED BY '"'
                        ESCAPED BY '\\' 
                        LINES TERMINATED BY '\r\n' ; 
                    """.format(sql,temp_path)

        if os.path.exists(temp_path):
            os.remove(temp_path)

        dump_sql = text(dump_sql)
        conn.execute(dump_sql)
        #获取到目标后，将结果复制到目标文件夹并且转换成EXCEL格式
        if save_path.split('.')[-1] == 'xlsx':
            result_df = pd.read_csv(temp_path)
            result_df.to_excel(save_path,header= header,index=False)
            os.remove(temp_path)
        else: #如果不是保存的XLSX格式，就直接复制CSV结果到目标文件夹
            copyfile(temp_path,save_path)     

        print(save_path,'数据已保存')
    else:
        enter_exit(how,"不能识别该导出方式")
        
    conn.close()
    db.dispose()

def insert_update_table(engine_text,df,table_name):
    """检查表的主键字段，根据主键字段，采用update的方式更新数据"""
    db = create_engine(engine_text,poolclass=NullPool)
    conn = db.connect()
    
    sql_unique_key = text("""
                        SELECT k.COLUMN_NAME
                        FROM information_schema.table_constraints t
                        LEFT JOIN information_schema.key_column_usage k
                        USING(constraint_name,table_schema,table_name)
                        WHERE t.constraint_type in ('UNIQUE','PRIMARY KEY') 
                            AND t.table_schema=DATABASE()
                            AND t.table_name='{0}';
        """.format(table_name))

    result = conn.execute(sql_unique_key).fetchall()

    #结果只有一列,去掉seq
    unique_columns = set([x.values()[0] for x in result if x.values()[0] !='seq'])

    if unique_columns :  #如果存在唯一性约束（包括主键约束）
        #需要更新的字段 去掉 唯一性约束字段 就是需要更新的字段
        update_columns = [x for x in df.columns if x not in unique_columns]
        #将数据写入临时表采用insert into on duplicate update的方式更新目的表
        sql_drop_temp = text("drop table if exists temp_insert;")
        
        sql_insert = text(""" create temporary table temp_insert as 
                              (select * from {0} );""".format(table_name))

        conn.execute(sql_drop_temp)
        conn.execute(sql_insert)

        df.to_sql('temp_insert',con=conn,if_exists='append',index=False)

        conn.execute(""" set sql_safe_updates = 0; """)

        sql_duplicate_update_list = [ ] 
        for u in update_columns: 
            sql_duplicate_update = """{0}.{1} = temp_insert.{1}""".format(table_name,u)
            sql_duplicate_update_list.append(sql_duplicate_update)

        sql_duplicate_update = ','.join(sql_duplicate_update_list)

        sql_insert_update =  text("""insert ignore into {0} 
                                   select * from temp_insert 
                                   on duplicate key update 
                                   {1};""".format(table_name,sql_duplicate_update))

        conn.execute(sql_insert_update)
        conn.execute("""drop table temp_insert;""")

    else: #如果不存在唯一性约束,直接写入
        df.to_sql(table_name,con=conn,if_exists='append',index=False)

    conn.close()
    db.dispose()


if __name__ == '__main__' :
    engine_text = gen_engine_text(r"E:\脚本\get_sql_data-master\connection_config.ini",'oracle+cx_oracle')

    sql = f' select plateSymbol, symbol from scrap_platesymbol  '
    fetch_result = execute_fetchall_engine(engine_text, sql)

    for row in fetch_result.iterrows():
        print(row[1])





import re 
from collections import defaultdict, Counter
from .regex_functions import replace_punctuations , replace_re_special, get_keyword_pat
from .os_functions import enter_exit
import xlrd 

def convert_key2list(word_dict):
    word_list = []
    for k, v in word_dict.items():
        for w in v :
            word_list.append(w)
    return word_list

def get_keyword_dict(path_list):
    #保存每个关键词列所需颜色的文字
    keyword_dict = defaultdict(set)
    #保存每个关键词列 类别的数字
    keyword_format_dict = defaultdict(str)

    if type(path_list) != list:
        path_list = [ path_list ]

    for path in path_list:
        wb = xlrd.open_workbook(path)
        #sheet name传入颜色
        sheet_names = wb.sheet_names()
        for sn in sheet_names:
            ws = wb.sheet_by_name(sn)
            #表头,根据表头获取应该写入红色还是蓝色，还是粗体
            header_list = []
            try:
                for x in ws.row(0):
                    if type(x.value) == str and x.value.strip() != '':
                        header = x.value.strip()
                    elif (type(x.value) == float or type(x.value) == int) :
                        header = str(x.value).rstrip('0').rstrip('.').strip()
                    else:
                        #为了防止两列中间隔一个空的表头单元格
                        header = None

                    if header != None:
                        header_list.append(header)

                if not header_list:
                    enter_exit(f'Error when reading keywords:\n{path}-"{sn}" should have at least one table header(keyword column names).')
            except IndexError:
                    enter_exit(f'Error when reading keywords:\n{path}-"{sn}" should have at least one table header(keyword column names).')

            seen_keywords = set()
            for row in list(ws.get_rows())[1:]:
                for i,format_word in enumerate(header_list):
                    if format_word != None:
                        keyword_value = row[i].value 
                        if type(keyword_value) == float and math.ceil(keyword_value) == keyword_value:
                            keyword = str(keyword_value).rstrip('0').rstrip('.').strip()
                        else:  #必须去掉容易导致歧义的特殊符号
                            keyword = replace_re_special(str(keyword_value).strip().lower())

                        if keyword not in seen_keywords and keyword != "" :
                            keyword_dict[format_word].add(keyword)

                            seen_keywords.add(keyword)

            #记录将每个颜色对应的关键词类
            for h in header_list:
                if h != None :
                    keyword_format_dict[h] = sn.strip().lower() 

        wb.release_resources()

    return keyword_dict, keyword_format_dict

def get_stopword_list(stopwords_path):

    stopword_list = defaultdict(int)

    with  open(stopwords_path,'r') as file:
        stopwords = file.read().splitlines() 

    for s in stopwords:
        if s.strip() != '':
            stopword_list[s.strip()] = 1

    return stopword_list

def process_text_eng(text, keyword_list=[], stopword_list=[], count_keywords_only = False):
    #仅适用英文
    
    #需要确保每个单词两边有空格
    keyword_list = [ ' ' + k + ' ' for k in keyword_list ]

    #加密邮件和数字
    text = encript_email_pat(text)
    text = encript_number_pat(text)

    text = replace_punctuations(text, replace_to_symbol=' ', exclude=['@']).lower().strip()

    text_list = text.split()
    #split空格，去掉多空格，再重新组合，为了匹配单词的两边空格，text两边也要加上空格
    text = ' ' + ' '.join(text_list) + ' '

    #keyword_dict先加入本身就是单词的词组，再组合成regex里面的格式拆分
    if not count_keywords_only and keyword_list:
        keyword_list = keyword_list + text_list

    #英文的处理，关键词搜索两边加上空格确保搜到的是英文
    keyword_pat = get_keyword_pat(keyword_list)

    if count_keywords_only:
        text_list = re.findall(keyword_pat, text, flags=re.I)
    else:
        text_list = re.split(keyword_pat,text,flags=re.I)
        text_list = [ x.lower() for x in text_list if x.strip() != '' ] 

    text_list = [ t.strip() for t in text_list ]

    if stopword_list:
        text_list = remove_stopwords(stopword_list, text_list)

    text_list = remove_numbers(text_list)
    text_list = remove_one_letter(text_list)

    text_list = [x.capitalize() for x in text_list]

    return text_list

def get_word_freq_dict(text, keyword_list, stopword_list, count_keywords_only=False, word_num=200):

    text_list = process_text_eng(text,keyword_list,stopword_list, count_keywords_only=count_keywords_only )

    word_count = dict(Counter(text_list).most_common(word_num))

    return word_count

def remove_stopwords(stopword_list, word_list):

    #转成字典再处理,尽量避免用 in list 的方式搜索
    stopword_dict = { s :1 for s in stopword_list }
    new_word_list = [ ]
    for w in word_list:
        if stopword_dict.get(w,None) == None:
            new_word_list.append(w.lower().strip())

    return new_word_list

def remove_numbers(word_list):
    word_list = [ x for x in word_list if x.isdigit() == False]
    return word_list




def remove_one_letter(word_list):
    word_list = [x for x in word_list if len(x) >= 2 ]
    return word_list

def encript_email_pat(text):

    if type(text) == str and text.strip() != '':
        email_like_pat = '([a-z0-9]{5,30})(@[^\u4e00-\u9fa5]+\.[a-z0-9]{2,15})'

        while True:
            if_match = re.search(email_like_pat, string=text)

            if if_match != None:
                text = re.sub(email_like_pat,repl='*****\g<2>',string=text)
            else:
                break

    return text

def encript_number_pat(text):
    if type(text) == str and text.strip() != '':
        result = ""

        number_counter = 0 

        new_text = str(text)
        for c in new_text:
            if c.isnumeric() == True:
                number_counter += 1
                if number_counter > 3 :
                    result += '*'
                else:
                    result += c
            else:
                number_counter = 0 
                result += c   
                
        #判断是否纯数字，如果是，后面需要rstrip('.0')
        if type(text) != str :
            return result.rstrip('.0')
        else:
            return result 
    else:
        return text



import os 
import re  
import time 
import sys 
import xlrd 
import datetime 
import codecs 
import hashlib
from pathlib import Path
from glob import glob 
from collections import defaultdict


def get_folder_list(path='.\\',folder_name = ''):

    if folder_name != '':
        folder_list = [ f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))==False\
                                 and folder_name.lower().strip() in f.lower() and f != '__pycache__' ]
    else:
        folder_list = [ f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))==False\
                                 and f != '__pycache__' ]

    return folder_list

def choose_folder(path=r'.\\', folder_name=''):
    """
    #参考choose_file 函数，除了选的是文件夹，其他地方都一样
    提供选项选择读取处理哪个EXCEL文档sheet,处理哪个字段,默认处理main_body
    """
    folder_list = get_folder_list(path, folder_name)

    folder_str_list = [' ' + str(i) +'-' + x for i,x in enumerate(folder_list)]

    if len(folder_list) == 0:
        enter_exit('No folder found!')
    elif len(folder_list) == 1 :
        folder_choose_index = 0
    else:
        print('\n')
        print('\n'.join(folder_str_list))
        folder_choose_index = input('\n---Please Select Folder Number(Default:First Folder): ').strip()
        #打开文档
        if folder_choose_index == '':
            folder_choose_index = 0

    choose_index_tag = 0 
    while choose_index_tag < 1:
        try:
            folder_choose_index = int(folder_choose_index)
            folder_choose = folder_list[folder_choose_index]
            choose_index_tag += 1 
        except :
            print('Invalid folder number!')
            folder_choose_index = input('\n---Please Select Folder Number(Default:First Folder): ').strip()

    return os.path.join(path,folder_choose)

def get_most_upper_level_path(file_name):
    #获取最高一个层级的目录位置，用来做临时文件储存地点
    cwd_upper_dir = Path(os.getcwd())
    cwd_most_upper_level = str(cwd_upper_dir).split('\\',1)[0]

    temp_path = os.path.join(f'{cwd_most_upper_level}',f'\\{file_name}')

    if os.path.exists(temp_path):
        os.remove(temp_path)    

    return temp_path

def generate_md5(string):
    return hashlib.md5(string.encode('utf-8')).hexdigest()

def last_day_of_month(any_day):
    next_month = any_day.replace(day=28) + datetime.timedelta(days=4)
    return next_month - datetime.timedelta(days=next_month.day)

def enter_exit(print_info=''):
    print(print_info)
    input('\nPress Enter to exit')
    sys.exit()

def partial_filename_path(path,partial_file_name,prefered_num=-1):
    """
    通过部分文件名定位到其中一份文件
    """
    file_pathes = [ x for x in os.listdir(path) if '~$' not in x ]
    file_pathes = [ x for x in file_pathes if partial_file_name in x]
    file_pathes = [os.path.join(path,x) for x in file_pathes]
    #排序
    file_pathes.sort()

    if file_pathes:
       #找到多个符合条件的文件优先取prefered_num的一个
        file_path = file_pathes[return_num]
    else:        
        file_path = None
    
    return file_path

def get_walk_files(path):
    """
    填入文件夹位置获取该目录和其子目录下的所有文件列表,获取的相对路径
    :param path 
    return a list of files inside the path
    """
    file_name_list = []
    for root,dirs,files in os.walk(path):
        for file  in files:
            file_name_list.append(file)
    file_name_list = [x for x in file_name_list if '~$' not in x]
    return file_name_list

def get_walk_abs_files(path):
    """
    填入文件夹位置获取该目录和其子目录下的所有文件列表,获取的绝对路径
    :param path 
    return a list of files inside the path
    """
    if not os.path.exists(path):
        enter_exit(f"Folder {path} doesn't exists !")
    file_name_list = []
    for root,dirs,files in os.walk(path):
        for file  in files:
            file_path = os.path.join(root,file)
            if os.path.isfile(file_path) and '~$' not in file :
                file_name_list.append(file_path)

    return file_name_list

def get_walk_folders(path):
    """
    填入文件夹路径获取该目录和其子目录下的所有文件夹列表
    :param path
    return a list of folders inside the path
    """
    folder_name_list = [ ]
    for root,dirs,files in os.walk(path):
        for d in dirs:
            folder_name_list.append(os.path.join(root,d))
    return folder_name_list

def check_require_files(path,require_name_list,regex=False):
    """
    检查某个路径是否包含必须的文档
    :param path:路径
    :param require_name_list: 要检查的文档/文件夹是否存在
    :param regex : 是否需要用re去匹配
    """
    path_files = get_walk_abs_files(path)
    #如果是精确匹配
    if regex == False:
        require_list = [ x for x in require_name_list if '~$' not in x and x in path_files]
    else: 
        require_dict = defaultdict(str)
        require_list = [ (x,re.search(x,y).group()) for x in require_name_list for y in path_files \
                         if re.search(x,y)!=None and '~$' not in  y ]
        for r in require_list:
            require_dict[r[0]] = r[1]
        require_list = require_dict.keys()
    #缺失的文件/文档
    file_lack = set(require_name_list) - set(require_list)
    if file_lack :
        print('请补充缺失的文档或文件夹:',','.join(file_lack))
        return False 
    else:
        return True

def get_require_files(path,require_file_list,regex=True,matched_part='xls',if_walk_path=True):
    """
    检查某个路径是否包含必须的文档
    :param path:路径
    :param require_file_list: 要检查的文档/文件夹是否存在
    :param regex : 是否需要用re去匹配
    :return : 如果不存在，返回空字典，如果文档存在 返回需要文档的对应绝对路径字典
    """
    if type(require_file_list) != list:
        require_file_list = [ require_file_list ]
        
    if if_walk_path == True:
        path_files = get_walk_abs_files(path)
    else: 
        path_files = [ os.path.join(path,x) for x in os.listdir(path) ]

    path_files.sort()

    require_dict = defaultdict(str)
    #如果是精确匹配
    if regex == False :
        require_list = [ (x,y) for x in require_file_list 
                               for y in path_files if '~$' not \
                               in  x and '.py' not in x and x == y.split('\\')[-1] ]  
    else: 
        if matched_part == None:
            require_list = [ (x,y) for x in require_file_list \
                                   for y in path_files if x.lower() in y.lower().split('\\')[-1] \
                                   and '~$' not in y ]
        else:
            require_list = [ (x,y) for x in require_file_list \
                                   for y in path_files if x.lower() in y.lower().split('\\')[-1] \
                                   and '~$' not in y and matched_part in y ]

    for r in require_list:
        if r[0] not in require_dict.keys():
            #记录返回的目标路径,已经有记录的不再更新
            require_dict[r[0]] = r[1]

    #对比哪些文档缺失
    require_list = require_dict.keys()
    #缺失的文件/文档
    file_lack = set(require_file_list) - set(require_list)

    if file_lack :
        enter_exit('Required files not found:{0}'.format(','.join(file_lack)))
        return {} 
    else:
        return require_dict

def get_require_file_list(path,require_file_list,regex=True,matched_part='xls',if_walk_path=True):
    """
    检查某个路径是否包含必须的文档
    :param path:路径
    :param require_file_list: 要检查的文档/文件夹是否存在
    :param regex : 是否需要用re去匹配
    :return : 如果不存在，返回空字典，如果文档存在 返回需要文档的对应绝对路径字典
    """
    if type(require_file_list) != list:
        require_file_list = [ require_file_list ]
        
    if if_walk_path == True:
        path_files = get_walk_abs_files(path)
    else: 
        path_files = [ os.path.join(path,x) for x in os.listdir(path) ]

    path_files.sort()

    require_dict = defaultdict(set)
    #如果是精确匹配
    if regex == False :
        require_list = [ (x,y) for x in require_file_list 
                               for y in path_files if '~$' not \
                               in  x and '.py' not in x and x == y.split('\\')[-1] ]  
    else: 
        if matched_part == None:
            require_list = [ (x,y) for x in require_file_list \
                                   for y in path_files if x.lower() in y.lower().split('\\')[-1] \
                                   and '~$' not in y ]
        else:
            require_list = [ (x,y) for x in require_file_list \
                                   for y in path_files if x.lower() in y.lower().split('\\')[-1] \
                                   and '~$' not in y and matched_part in y ]

    for r in require_list:
        #记录返回的目标路径,已经有记录的不再更新
        require_dict[r[0]].add(r[1])

    #对比哪些文档缺失
    require_list = require_dict.keys()

    #缺失的文件/文档
    file_lack = set(require_file_list) - set(require_list)

    if file_lack :
        enter_exit('Required files not found:{0}'.format(','.join(file_lack)))
        return False 
    else:
        return require_dict

def check_create_new_folder(dir_or_path):
    """
    检查是否存在某个文件夹，或者输入的文件所在的目录是否存在，如果不存在话 就生成新的文件夹
    """
    path = Path(dir_or_path)
    if not path.exists():
        path.mkdir(mode=0o777,parents=True,exist_ok=True)
    #如果传入的是一个文档，需要删掉新建的以文档命名的文件夹
        if not path.is_file() :
            path.rmdir()
    return dir_or_path

def choose_file(path=r'.\\'):
    """
    提供选项选择读取处理哪个EXCEL文档sheet,处理哪个字段,默认处理main_body
    """
    file_list = [ x for x in os.listdir(path) if '~$' not in x and '.xls' in x ]
    file_str_list = [' ' + str(i) +'-' + x for i,x in enumerate(file_list)]

    if len(file_list) == 0:
        enter_exit('Folder is empty')
    elif len(file_list) == 1 :
        file_choose_index = 0
    else:
        print('\n')
        print('\n'.join(file_str_list))
        file_choose_index = input('\n---Please Select File Number(Default:First File): ').strip()
        #打开文档
        if file_choose_index == '':
            file_choose_index = 0

    choose_index_tag = 0 
    while choose_index_tag < 1:
        try:
            file_choose_index = int(file_choose_index)
            file_choose = file_list[file_choose_index]
            choose_index_tag += 1 
        except :
            print('Invalid file number!')
            file_choose_index = input('\n---Please Select File Number(Default:First File): ').strip()

    return os.path.join(path,file_choose)

def choose_sheet_column(file_path):
    """
    打开文档选择sheeth和column,并且返回一个worksheet内容,方便后面使用
    """
    print('\nReading Sheet...')
    wb = xlrd.open_workbook(file_path, on_demand=True)
    sheet_names = wb.sheet_names()
    sheet_str_list = [ ' ' + str(i) +'-' + str(x).rstrip('0').rstrip('.') for i,x in enumerate(sheet_names) ]

    if len(sheet_names) == 1 :
        sheet_choose_index = 0
    else:
        print('\n'.join(sheet_str_list))
        sheet_choose_index = input('\n---Please input Sheet Number and press Enter(Default:First sheet): ').strip()
        #打开文档
        if sheet_choose_index == '':
            sheet_choose_index = 0

    choose_index_tag = 0 
    while choose_index_tag < 1:
        try:
            sheet_choose_index = int(sheet_choose_index)
            sheet_choose = sheet_names[sheet_choose_index]
            choose_index_tag += 1 
        except :
            print('Invalid Sheet number!')
            sheet_choose_index = input('\n---Please input Sheet Number and press Enter(Default:First sheet): ').strip()

    #打开工作簿 读取列
    ws = wb.sheet_by_name(sheet_choose)
    columns = [x.value for x in ws.row(0)]

    column_str_list = [ ' ' + str(i) +'-' + str(x).rstrip('0').rstrip('.') for i,x in enumerate(columns)]

    if len(columns) == 1 :
        column_choose_index = 0
    else:
        print('\n')
        print('\n'.join(column_str_list))
        column_choose_index = input('\n---Please input the Column Number that needs highlighted（Default:main_body）:').strip()
        #打开文档
        if column_choose_index == '':
            try:
                column_choose_index = columns.index('main_body')
            except ValueError:
                column_choose_index = 0

    column_choose_index = int(column_choose_index)
    column_choose = columns[column_choose_index]    
    #返回的是xlrd的wb
    return wb,ws,column_choose



from .sql_functions import execute_fetchall_engine
from .excel_functions import save_xlsxwriter_wb
from .os_functions import enter_exit
from xlsxwriter.exceptions import FileCreateError
import re 
import os 
import pandas as pd 
import time 
import configparser

#读取ini配置
cfparser = configparser.ConfigParser()
cfparser.read('mysql_connection_config.ini',encoding="utf-8")

config_sections = cfparser.sections()

host = cfparser.get('connection','host')
port = cfparser.get('connection','port')
database = cfparser.get('connection','database')
charset = cfparser.get('connection','charset')
username = cfparser.get('connection','username')
password = cfparser.get('connection','password')

#读取每次超过50W行就保存到另一个XLSX文档
seperate_batch = 500000

engine_text = f'mysql://{username}:{password}@{host}:{port}/{database}?charset={charset}'

with open('query_sql.txt','r',encoding='utf-8') as file:
	sql = file.read()
	#去掉可能出现的看不见的UTF8-BOM空格
	sql = sql.replace('\ufeff','')

t1 = time.clock()
df = execute_fetchall_engine(engine_text=engine_text,sql=sql)

t2 = time.clock()

print('Results get in',round(t2 - t1,0),'seconds')

df_length = df.shape[0]
save_name = 'sql_result'

save_path = "{0}.xlsx".format(save_name)

print('Writing to excel')

if df_length > seperate_batch:
	counter = 0 
	for i in range(0,df_length,seperate_batch):
		counter += 1 
		df_batch = df.iloc[i:i+seperate_batch,:]

		print('Getting the {}th Row to {}th Row'.format(i,i+seperate_batch))
		#创建新的文档
		save_path  = "{0}_{1}.xlsx".format(save_name,counter)

		writer = pd.ExcelWriter(save_path,engine='xlsxwriter',options={'strings_to_urls': False,\
						'strings_to_formulas': False},date_format='yyyy-mm-dd',datetime_format='yyyy-mm-dd')

		df_batch.to_excel(writer,sheet_name='result',index=False)

		save_xlsxwriter_wb(writer,save_path)
		t3 = time.clock()
		print(round(t3-t2,0),'seconds used')

else:#如果并没有超过预定的记录数直接提取后写入 
	writer = pd.ExcelWriter(save_path,engine='xlsxwriter',options={'strings_to_urls': False,\
					'strings_to_formulas': False},date_format='yyyy-mm-dd',datetime_format='yyyy-mm-dd')

	df.to_excel(writer,'result',index=False)
	save_xlsxwriter_wb(writer,save_path)
	t3 = time.clock()
	print(round(t3-t2),'seconds used')

enter_exit()


import re 
from collections import defaultdict 
from string import punctuation
import string 

from flashtext import KeywordProcessor

# keyword_processor = KeywordProcessor()
# for i in ['tests','testss','test','5G is(not','ok','100%','不可能吧','优势','在哪里','哪里']:
#     keyword_processor.add_keyword(i)

# print(keyword_processor.get_all_keywords())

# text = 'tests  100% are  do.ne testss/5G is(not ok'
# # text = '你觉得5G或者优势在哪里'
# kw_found = keyword_processor.extract_keywords(text)

# print(kw_found)

# exit()

def get_keyword_pat(keyword_list):
    keyword_list = sorted(set(keyword_list), key=len, reverse=True)

    keyword_pat = u'('+ '|'.join(keyword_list) + ')' 

    return keyword_pat

#数字模糊匹配函数,数字连着命中才算符合条件，即128对应的必须是128才算命中一个，命中12不算
def number_similarity(a,b, common_pattern):
    #该函数需要传入两者同时共存的内存模式,如果不填就简单地判断两个列表存在相同的匹配到的数字即可(y3 16G+128)可能会匹配到(3+16G)
    if common_pattern != '':
        a_match = re.search(common_pattern, a, flags=re.I)
        if a_match == None:
            return [ ]
        else:
            b_match = re.search(common_pattern, b, flags=re.I)
            #b是标准结果
            number_list_a = re.findall('\d+',a_match.group())
            number_list_b = re.findall('\d+',b_match.group())

            if len(number_list_a) > 0  :
                intersection = set(number_list_a) & set(number_list_b)
                if intersection and  set(number_list_b) == intersection:
                    return number_list_b

    return [ ] 
    
def check_syn_str_regex(string_a, string_b,special_syn_list):
    string_a = string_a.lower()
    string_b = string_b.lower()

    match_result_a = [ re.search(x, string_a, flags=re.I).group() for x in special_syn_list if re.search(x, string_a, flags=re.I) != None ]  
    match_result_b = [ re.search(x, string_b, flags=re.I).group() for x in special_syn_list if re.search(x, string_b, flags=re.I) != None ]

    intersection = set(match_result_a) & set(match_result_b)

    return len(intersection) == len(match_result_a)


def check_syn_str_regex_number(string_a, string_b, special_syn_list):
    #数字的不需要判断intersection,确认两者同时存在即可,即两者都存在 
    #x = check_syn_str_regex_number('(128G+8G)','8+28G',['(\d{1,4}[GB]?\+\d{1,4}[GB]?)|(\d{1,4}G{1}B?)']) -->True因为都符合
    string_a = string_a.lower()
    string_b = string_b.lower()

    match_result_a = [ True if re.search(x, string_a, flags=re.I) != None else False for x in special_syn_list ]  
    match_result_b = [ True if re.search(x, string_b, flags=re.I) != None else False for x in special_syn_list ]  

    if len(match_result_a) != len(match_result_b):
        return False 
    else:
        for a, b in zip(match_result_a,match_result_b):
            if a == True and b == True :
                return True
    return False


def strB2Q(ustring):
    """把字符串半角转全角"""
    ss = []
    for s in ustring:
        rstring = ""
        for uchar in s:
            inside_code = ord(uchar)
            # 全角空格直接转换
            if inside_code == 32:  
                inside_code = 12288
            # 全角字符（除空格）根据关系转化
            elif (inside_code >= 33 and inside_code <= 126):  
                inside_code += 65248
            rstring += chr(inside_code)
        ss.append(rstring)
    #顿号要转成逗号
    return ''.join(ss)


def strQ2B(ustring):
    """把字符串全角转半角"""
    ss = []
    for s in ustring:
        rstring = ""
        for uchar in s:
            inside_code = ord(uchar)
            if inside_code == 12288:  # 全角空格直接转换
                inside_code = 32
            elif (inside_code >= 65281 and inside_code <= 65374):  # 全角字符（除空格）根据关系转化
                inside_code -= 65248
            rstring += chr(inside_code)
        ss.append(rstring)

    return ''.join(ss)

def strQ2B(ustring):
    """把字符串全角转半角"""
    halfwidth_symbol = '!\"\"#$%&\'\'()*+,-./:;<=>?@[]_{|}~ '
    fullwidth_symbol = '！“”#$%&‘’（）*+，-。/：；《=》？@【】_{|}~ '
    translator = str.maketrans(fullwidth_symbol, halfwidth_symbol)

    ustring = ustring.translate(translator)
    return ustring


def strip_puntuations(input_string):
    #清空字符串两边的所有符号
    for x in punctuation:
        input_string = input_string.strip(x)
    return input_string

def replace_punctuations(input_string, replace_to_symbol=' ', exclude = [ ] ):
    #将字符串中的所有符号统一替换成空格
    if type(exclude) != list:
        exclude = [ exclude ]

    if type(input_string) != str:
        return input_string

    mapping_result = ''.join([replace_to_symbol if s not in exclude else s for s in string.punctuation])

    translator = str.maketrans(string.punctuation, mapping_result)
    input_string = input_string.translate(translator)
    return input_string 

def replace_multi_symbol(string, symbol):
    """把多个符号替换成单个，比如多个换行符 替换成 一个换行符,replace('\n\n','\n')并不能解决问题"""
    symbol_double = symbol + symbol
    while symbol_double in string:
        string = string.replace(symbol_double,symbol)
    return string

def symbol_to_spaces(string):
    string = replace_multi_symbol(replace_punctuations(strQ2B(string)).strip(),' ')
    return string

def normalize_punctuations(string):
    #用来做拆分（匹配是需要把符号全变成空格，拆分不能改变结果符号）
    #把string的所有符号标准化（全角全部统一转成半角，连续重复的符号变成单个,两边都不留空格或者换行符）
    #转成半角
    # string = strQ2B(string)
    # for s in punctuation:
    #     string = replace_multi_symbol(string, s)
    #     #确保两边没有特殊符号
    # string = string.strip()
    return string   

def replace_re_special(word):
    #注意\要写在前面，因为后面循环替换了\\进去
    for special_symbol in r'\-+()[]{}.*^$~|?^,':
        new_special_symbol = '\\' + special_symbol
        word = word.replace(special_symbol, new_special_symbol)
    return word


def search_en(combine_string):
    """提取出中英文混合字符串中的英文部分"""
    non_cn_pat = "[^\u4e00-\u9fa5]+"
    en_pat = ".*(\w)+.*"
    found_all = re.findall(non_cn_pat,combine_string)
    en_found_all = []

    if found_all :   #定位有英文的部分
        en_found_all = [re.search(en_pat,x).group() for x in found_all if re.search(en_pat,x) != None]

    if en_found_all :
        return en_found_all[0]
    return None

def partial_match_pct(short_str,long_str,special_syn=['pro','plus','max','youth']):
    """
    short_str : original string
    long_str : standard target string 
    匹配机型，不要放入两个一样的字符串，获取短字符串在长字符串中是否存在，并且占了多少个字符,
    从开头开始匹配,不替换空格不用in的方式查找，带有括号的机型匹配优先级最高，
    通常Y15S和Y15(4+64G)后者更容易被缩写成Y15"""
    #20200107 :检查长字符串是否包含有特殊字符
    def check_syn_str(new_str,special_syn):
        new_str = new_str.lower()
        match_syn_list = [x for x in special_syn if x in new_str]
        if match_syn_list:
            return match_syn_list[0]
        else:
            return 'NA'

    #暂时不懂怎么把IQOO 3 4G 优先匹配到 IQOO 3
    def match_sequence_str(short_string,long_string):
        #获取一个字符串对应另一个字符串匹配的字母
        # input 'iQOO 3 4G', IQOO 3 (5G)  --> IQOO 3
        # input 'iQOO 3 4G', IQOO 3 (4G)  --> IQOO 3 4
        #复制相同的一个
        long_counter = 0 
        temp_record = ''

        for s1 in short_string:
            for s2 in long_string[long_counter:] :
                if s1 == s2:
                    long_counter += 1 
                    temp_record += s2
                    break
        return temp_record 


    default_result = (0,long_str)

    if type(short_str) != str:
        return default_result

    short_str = short_str.strip().lower().replace('\n','').replace('  ',' ')
    new_long_str = long_str.strip().lower().replace('\n','').replace('  ',' ')

    #去掉换行符和多空格之后相等的话 直接返回长字符串
    if short_str == new_long_str:
        return (99,long_str)
    # #防止放入同一字符串
    # if short_str == new_long_str: 
    #   return default_result
    #括号和空格都要分割处理
    if '(' in new_long_str or '（' in new_long_str :
        new_long_str = new_long_str.replace('（','(').split('(')[0]
    # elif ' ' in new_long_str :
    #   new_long_str = new_long_str.split()[0]
    """ 匹配可能包含错误拼写，包括漏写，多写，错写的机型名, 机型名一般有NEX, S1， X3S, V1Max, V15Pro,
    允许的错别字条件是：不允许数字写错，不允许前面的字母写错，当字母大于等于3个时，允许漏写或者错写，多写2个字母,
    比如pro写成pr ,prou, max写成ma, V15Pro 写成 V15P（如果有V15P应该在之前就可以关联上，所以排除他是V15P的可能，
    更大可能是想写V15Pro）"""
    #从头开始匹配的时候，如果完整的short_str是准确拼写的，正常返回，如果有错别字，采用else以下的匹配方式
    match_short_str = ''
    #前两/三位字母相同,方便后面elif写条件
    first2_letter_equal = short_str[:2] == new_long_str[:3].split(' ')[0][:2]
    first3_letter_equal = short_str[:3] == new_long_str[:3].split(' ')[0][:3]
    #前3位字符串包含2个数字的情况 必须满足前3个字母相同,
    first3_letter_2num_short = re.search('[0-9]{2,}',short_str[:3]) != None
    first3_letter_2num_long =  re.search('[0-9]{2,}',new_long_str[:3]) != None

    first3_letter_less2num_short = not first3_letter_2num_short
    first3_letter_less2num_long = not first3_letter_2num_long

    #一个字符直接返回0,两个字符，去掉空格之后的前两个字符要完全相等,并且要确保长字符串里没有special_syn的字符
    if len(short_str) == 2 :  
        if not check_syn_str(new_long_str,special_syn):
            new_long_str = new_long_str.split(' ')[0].split('(')[0]
            if short_str == new_long_str:
                return (2/len(long_str),long_str)
            else:
                return default_result
        else:  #如果长字符串含有pro等字符 直接判断不匹配
            return default_result
    #至少出现3个字符,并且前两个字母(如果前两位是字母+数字，后面不再有数字),如果前3位包含2个数字,前3个字符要相同,规避掉V 15这种空格的情况 
    elif (len(short_str) >= 3 and first2_letter_equal) and \
         (
               (first3_letter_2num_short and first3_letter_2num_long and first3_letter_equal) \
            or (first3_letter_less2num_short and first3_letter_less2num_long and first2_letter_equal)
         ):

        for i in range(2,len(short_str)+1) :
            if short_str[:i].replace(' ','') in new_long_str.replace(' ',''):
                match_short_str = short_str[:i]
                continue        
        #优先计算匹配率大的字符串,并且为了实现区分V11i 优先匹配到V11而不是V11Pro的情况，而外加一个长字符串的比率
        #--计算结果相同(0.75, 'V11Pro')  (0.75, 'V11'),后面的sort比较难实现long_str的顺序排列
        if ' ' in long_str or '(' in long_str:
            long_str_bias = len(match_short_str)/len(long_str.split(' ')[0].split('(')[0])/100  #比例需要调小
            #如果去掉空格和( 符号之后的long_str_bias仍然相等，就将原来的标准机型全部去掉这些字符，对比整体全部的匹配度，做一个bias
            # long_str_bias += len(match_sequence_str(short_str,long_string)) / (len(short_str) + len(long_string)) / 1000  #比例更小
        else:  #没有出现空格和（ 的 不带pro的应该优先
            long_str_bias = len(match_short_str)/len(long_str)/100 + 0.00001 

        #确保短字符和长字符同时有或者同时没有special_syn_str 
        if check_syn_str(short_str,special_syn) == check_syn_str(new_long_str,special_syn):
            return (len(match_short_str)/len(short_str) + long_str_bias ,long_str)

    return default_result


def re_sub_replace(pattern,string,repace_symbols=('-','_'),group_name=1):
    """
    当re.sub(pattern,repl,string)内置的repl = "g<1>" 不能满足替换需求的时候,
    比如需要将group目标内的文字中的某个符号替换掉, 使用的时候要注意替换代码内的sub replace符号
    :param pattern : re pattern 
    :param string : original string 
    """
    def replace_match(matched):
        #re.sub会自动传入matched结果
        original_string = matched.group()
        #pattern支持填入group_name
        matched_string = matched.group(group_name)
        replace_string = original_string\
        .replace(matched_string,matched_string.replace(replace_symbols[0],replace_symbols[1]))
        return replace_string
    
    new_string = re.sub(pattern,replace_match,string,repace_symbols=('-','_'))
    return new_string


def re_findall_replace(pattern,string,replace_symbols=('-','_')):
    """
    当re.sub(pattern,repl,string)内置的repl = "g<1>" 不能满足替换需求的时候,
    比如需要将group目标内的文字中的某个符号替换掉,并且需要match多个group目标
    """
    matched_list = re.findall(pattern,string)
    new_string = string
    for mat in matched_list:
        new_string = new_string.replace(mat,mat.replace(replace_symbols[0],replace_symbols[1]))
    return new_string


def re_findall_sub(pattern,string,replace_symbols=('-','_')):
    """
    当re.sub(pattern,repl,string)内置的repl = "g<1>" 不能满足替换需求的时候,
    比如
    """
    matched_list = re.findall(pattern,string)
    new_string = string
    for mat in matched_list:
        new_string = new_string.replace(mat,mat.replace(replace_symbols[0],replace_symbols[1]))
    return new_string


def split_wrong_combine(pattern,sub_pattern,string):
    """
    需要处理爬虫换行符的问题, 抓取的时候把换行符去掉了，导致单词连接成错误的形式
    比如 axxBx, AxxxCONs, usagePROgreat, slow3.Great sunglass
     (注意 3app , 3PlayStore只能算是拼写错误，不需要拆分, LED3 拆不了，陷入无限循环)
    需要将两个错误合并的单词用换行符替换拆分,
    :param pattern : the original pattern that we want to find out
    :param sub_pattern : sub pattern to extract from pattern, 
    will be replaced with original + '\n' or '\n' + original  
    :param string : target string
    e.g
    pattern = '[A-Z]?[a-z]+[A-Z0-9]+[a-zA-Z]*' 
    sub_pattern = '[A-Z0-9]+[a-z]*'
    """
    #记录需要修改的部分
    new_string = string
    new_string_dict = defaultdict(str)
    matched_list = re.findall(pattern,new_string)
    if matched_list :
        for mat in matched_list:
            match_sub = re.search(sub_pattern,mat)
            #需要确保sub_pattern 能匹配出 pattern 匹配到的内容的部分目标
            if match_sub != None:
                match_sub = match_sub.group()
                replace_match = mat.replace(match_sub,'\n' + match_sub)
                #如果替换的是第一个单词，需要特殊处理。不要替换前面的符号为换行符，而是需要保持原来单词自带的 “空”
                #并且换行符 应该是加在第一个单词后面
                if [ x for x in replace_match].index('\n')  == 0:
                    replace_match = replace_match.replace('\n','')
                    replace_match  = mat.replace(match_sub, match_sub + '\n')

                replace_matched = split_wrong_combine(pattern,sub_pattern,replace_match)
                new_string_dict[mat] = replace_matched 
    else:
        return new_string

    for k,v in new_string_dict.items():
        new_string = new_string.replace(k,v)
        
    return new_string



#++++++++++++++++++以下废弃函数++++++++++++++++++++++++

# def standardize_country_by_cn_similarty(short_str, standard_str_list):
#     #处理中文国家缩写和完整国家名称无法匹配到的情况
#     standard_str_list = list(set([str(x).strip() for x in standard_str_list]))

#     standard_str_list = sorted(standard_str_list, key=len, reverse= False)
#     #通过前面字符串匹配 马来 -- > 马来西亚
#     standard_similarity_list = [ (s,1) if short_str in s else (s,0) for s in standard_str_list ]
#     if standard_similarity_list[0][1] > 0 :
#         return standard_similarity_list
#     else:
#         standard_similarity_list = [ ]
#         for ss in standard_str_list:
#             short_match_counter = 0
#             for ss_each_letter in ss:
#                 for s in short_str:
#                     if s == ss_each_letter:
#                         short_match_counter += 1

#             #至少能匹配上两个字
#             str_similarity =  short_match_counter / len(short_str) + short_match_counter / min([len(short_str),len(ss)])
#             if short_match_counter >= 2 :
#                 standard_similarity_list.append([ss,str_similarity])
#             else:
#                 standard_similarity_list.append([ss, 0 ])

#         standard_similarity_list = sorted(standard_similarity_list, key=lambda x:x[1],reverse=True)

#     return standard_similarity_list




"""
list , dict , itertools, Counter functions  
"""
import os
import re  
from itertools import combinations 
from collections import defaultdict


#检查sheet_name，如果出现重复会无法写入，后面加入不同的数字
def duplicate_elem_add_seq(lst):
	#给列表出现重复的字符串末尾加上序号
	temp_dict = defaultdict(int)
	result_list = [ ]
	for s in lst:
		if temp_dict[s] != 0:
			result_list.append(str(s) + str(temp_dict[s]))
		else:
			result_list.append(s)

		temp_dict[s] += 1 
		
	return result_list


#模糊匹配的函数
def lcs(a, b):
	# longest common subsequence
    tbl = [[0 for _ in range(len(b) + 1)] for _ in range(len(a) + 1)]
    for i, x in enumerate(a):
        for j, y in enumerate(b):
            tbl[i + 1][j + 1] = tbl[i][j] + 1 if str(x) == str(y) else max(tbl[i + 1][j], tbl[i][j + 1])
    res = []
    i, j = len(a), len(b)
    while i and j:
        if tbl[i][j] == tbl[i - 1][j]:
            i -= 1
        elif tbl[i][j] == tbl[i][j - 1]:
            j -= 1
        else:
            res.append(a[i - 1])
            i -= 1
            j -= 1
    return res[::-1]

#严格模糊匹配函数
def filter_lcs(a,b):
    #过滤掉 断层的命中字符串
    match_list = lcs(a,b)
    a = a.replace(' ','')
    b = b.replace(' ','')
    match_list = [ x for x in match_list if x != ' ']
    new_match_list = [ ]

    if match_list:
        for i in range(len(match_list)):
            if match_list[i] != a[i]:
                break 
            else:
                new_match_list.append(match_list[i])
    else :
        return new_match_list

    return new_match_list



def list_diff_outer_join(split_result, findall_result):
	#寻找没有匹配上的集合，求两个列表的outer join差集
	not_match_list = [ ]
	for elem_f in findall_result:
		while split_result :
			elem_s = split_result.pop(0)
			if elem_s != elem_f :
				not_match_list.append(elem_s)
			else :
				break
				
	return [ x for x in not_match_list + split_result if x != '' ] 

def sublist_combinations_all(value_list):
	"""获取一个完整列表的所有有序子列表的组成可能"""
	value_len = len(value_list)
	value_combination_list = [ ]
	for i in range(2,value_len+1): 
		value_combination = combinations(value_list,i)
		for c in value_combination:
			value_combination_list.append(list(c))

	return value_combination_list

def find_sublists_seq(value_list):
	"""获取一个完整列表的所有有序子列表的组成可能,
	[3, 4, 3, 5, 6] -->
	[[3, 4], 3, 5, 6]
	[3, [4, 3], 5, 6]
	[3, 4, [3, 5], 6]
	[3, 4, 3, [5, 6]]
	例如：not up to mark  notuptomark，not good --> notgood  """
	list_len = len(value_list)
	record_list = [ ]
	for word_num in range(2,list_len+1):  
		for start_pos in range(0,list_len - word_num + 1 ):
			temp_list =  value_list[:start_pos] \
					  + [value_list[start_pos:start_pos+word_num]] \
					  +  value_list[start_pos+word_num:]

			record_list.append(temp_list)
	return record_list

def find_sublist_indexes(complete_list,sub_list):
	"""从完整大列表中搜索出小的子列表 从头到尾 在大列表中对应的索引（优先第一个元素,可以求出前中后的所有索引位置
	[a,b,c,d,e] --> [b,d,e]  -->  [1,3,4]  --> result range_index_dict
	:param complete_list:
	:param sub_list
	:return : 	
	 range_index_dict = {
	'sub_list_index':record_index_list,  
	'pre_range_index':pre_range_index_list,
	'mid_range_index':mid_range_index_list,
	'back_range_index':back_range_index_list
	}
	"""
	#防止原始列表被修改，复制版本列表会被修改
	complete_list_copy = complete_list.copy()
	record_index_list = [ ]
	record_index = 0 
	for s in sub_list:
		split_index = complete_list_copy.index(s)
		#跳过本元素取下一个开始的列表后部分
		complete_list_copy = complete_list_copy[split_index + 1 : ]
		#记录每次的索引位置,split_index是每次拆分后的单独记录
		record_index += split_index 
		record_index_list.append(record_index)
		#每次带走一个元素,记录索引需要对应加上带走的元素数量(+1)
		record_index += 1 
	#重叠的元素范围 range_index_dict

	#返回前中后三个部分的所有概念index
	complete_list_index = [ x for x in range(len(complete_list))]
	pre_range_index_list = complete_list_index[:record_index_list[0]]
	mid_range_index_list = [ x for x in complete_list_index[record_index_list[0]:record_index_list[-1]+1] \
						if x not in record_index_list ] 

	back_range_index_list = complete_list_index[record_index_list[-1]+1:]

	range_index_dict = {
	'sub_list_index':record_index_list,  
	'pre_range_index':pre_range_index_list,
	'mid_range_index':mid_range_index_list,
	'back_range_index':back_range_index_list
	}

	return range_index_dict


# a = find_sublist_indexes(['C_FACE', 'D_BE_VERB', 'D_BAD_1', 'D_AND', 'D_WORK', 'D_PREPOSITION'],['C_FACE', 'D_BE_VERB', 'D_BAD_1'])

# print(a)

def convert_twolist2dict(list_a,list_b):
	"""
	将两个列表转成字典形式
	"""
	return dict(zip(list_a,list_b))

def list_transpose(original_list):
	"""列表转置"""
	return list(map(original_list,zip(*l)))



def uneven_list_dictify(values):
	#为了让最后一级也全都是键值,补充进空字符串
	values = [ x+ [''] for x in values ]
	d = {}
	for row in values:
		here = d
		for elem in row[:-2]:
			if elem not in here:
				here[elem] = {}
			here = here[elem]
		here[row[-2]] = row[-1]
	return d

def forward_fill(values):
	#区别于pandas的forwordfill,不会填充后面为空的部分，只向前填充前面列空的部分
	complete_row = values[0]
	result = [complete_row]
	#第一行必定是有完整的前几列数据,从第二行开始读取填充
	for v_index in range(1,len(values)):
		value = values[v_index] 
		for i in range(len(value)):
			#如果发现有缺失的部分
			if value[i] == '':
				value[i] = complete_row[i]
			else:
				complete_row = value
				break
		result.append(value)
	return result


def recursive_find_nodes(dictionary,parrent_node,record_list):
	#记录数字的previous_num为列表形式 [ ]
	#需要返回一组列表，形式为 [ current_node, parrent_node]
	#例子：# recursive_find_nodes(nested_dict,parrent_node = '',record_list= record_list)
	if type(dictionary) == dict:
		for k in dictionary.keys():
			recursive_find_nodes(dictionary[k],k,record_list)
			record_list.append([k,parrent_node])
	else:
		record_list.append([dictionary,parrent_node])

	return record_list

def recursive_gen_nodecodes(dictionary,parrent_node,record_list,level,pre_counter):
	#同recursive_gen_nodecodes, 添加了
	#不生成编码的函数
	#例子：recursive_gen_nodecodes(nested_dict,parrent_node = '',record_list= record_list,level=0,pre_counter=['00','00','00','00'])
	counter_list = [ ]  
	counter = 0 
	if type(dictionary) == dict:
		for k in dictionary.keys():
			counter += 1 
			pre_counter[level] = str(counter).zfill(2)
			record_list.append([''.join(pre_counter[:level+1]),k,parrent_node])
			recursive_gen_nodecodes(dictionary[k], k ,record_list,level + 1, pre_counter)
	else:
		record_list.append([''.join(pre_counter),dictionary,parrent_node])

	return record_list


from nltk.corpus import words as nltk_words
from collections import defaultdict

def gen_valid_wordlist(valid_word_len):
	"""
	创建有效的单词列表,全部返回小写
	"""
	word_list = [ x.lower() for x in nltk_words.words() if len(x) >= valid_word_len ]
	return word_list 


def convert_list2dict(word_list):
	"""
	把列表写成字典
	"""
	word_dict = defaultdict(int)
	for w in word_list:
		word_dict[w] = 1
	return word_dict





import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="common_utils_data",
    version="0.0.9",
    author="Tracy Tang",
    author_email="tracytang58@icloud.com",
    description="Fix bugs in the wrong imports",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Mi524/common_utils_pkg",
    project_urls={
        "Bug Tracker": "https://github.com/Mi524/common_utils_pkg/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
      'pandas>=1.2.0',
      'numpy',
      'openpyxl',
      'xlsxwriter',
      'xlwings',
      'html5lib',
      'lxml',
      'sqlalchemy>=1.3.22',
      'pymysql>=1.0.2',
      'mysql-client',
      'flashtext',
      'swifter>=1.0.7',
      'cx_Oracle'
  ],
)